//
// Created by nrjn on 15.11.20.
//

#include "ntc.h"

#include <drivers/gpio.h>
#include <drivers/adc.h>
#include <hal/nrf_saadc.h>
#include <math.h>


/* Device ADC - Voltage divider circuit parameters */
/* Caliberate this for every thermistor
#define PT1000_CALIBIRATION_OFFSET    4.0f
#define ADC_INTERNAL_REF_VOLTAGE      0.6f   // Volts
#define ADC_INPUT_VOLT_RANGE          3.3f   // Volts
*/
/*
 * P0.02 - AIN0
 * P0.03 - AIN1
 * P0.04 - AIN2
 * P0.05 - AIN3
 * P0.28 - AIN4
 * P0.29 - AIN5
 * P0.30 - AIN6
 * P0.31 - AIN7
 */

#define ADC_RESOLUTION                12
#define ADC_VALUE_RANGE_12_BIT        4095   // 2^12 - 1
#define ADC_GAIN                      ADC_GAIN_1_6
#define ADC_REFERENCE                 ADC_REF_INTERNAL
#define ADC_ACQUISITION_TIME          ADC_ACQ_TIME(ADC_ACQ_TIME_MICROSECONDS, 10)
#define ADC_1ST_CHANNEL_ID            5
#define ADC_1ST_CHANNEL_INPUT         NRF_SAADC_INPUT_AIN5

#define RESISTANCE_REFERENCE          100 * 1000
#define KELVIN_TO_CELSIUS             273.15
#define THERMISTOR_NOMINAL            100000.0
#define TEMPERATURE_NOMINAL           25
#define B_CONSTANT                    4550
#define ADC_SAMPLE_VALUE_MAX          10

#define ADC_DEV                       "ADC_0"
#define GPIO_DEV                      "GPIO_0"


static struct device *adc_dev;
static struct device *gpio_dev;

static int16_t resistance_to_temperature(float resistance);

int16_t resistance_to_temperature(float resistance)
/* for NCP03WF104F05RL (100k), B-constant (K) 25/50 C is 4250±1%  */
{
    double temperature_c = resistance / THERMISTOR_NOMINAL;    // (R/Ro)
    temperature_c = log(temperature_c);            // ln(R/Ro)
    temperature_c /= B_CONSTANT;            // 1/B * ln(R/Ro)
    temperature_c += 1.0 / (TEMPERATURE_NOMINAL + KELVIN_TO_CELSIUS);    // + (1/To)
    temperature_c = 1.0 / temperature_c;            // Invert
    temperature_c -= KELVIN_TO_CELSIUS;             // convert to C
    return (int16_t) (temperature_c);
}

int get_temperature(int16_t *temperature) {
    int err = 0;
    int16_t m_sample_buffer = 0;
    int32_t adc_avg = 0;

    if (!adc_dev) {
        return -1;
    }

    const struct adc_sequence sequence = {
            .channels = BIT(ADC_1ST_CHANNEL_ID),
            .buffer = &m_sample_buffer,
            .buffer_size = sizeof(m_sample_buffer), // in bytes!
            .resolution = ADC_RESOLUTION,
    };

    for (int i = 0; i < ADC_SAMPLE_VALUE_MAX; i++) {
        err = adc_read(adc_dev, &sequence);
        adc_avg += m_sample_buffer;
        k_msleep(50);
    }
    adc_avg /= ADC_SAMPLE_VALUE_MAX;

    // R = R_ref/ (1023/ADC - 1)
    float r = RESISTANCE_REFERENCE / (float) (ADC_VALUE_RANGE_12_BIT / (float) adc_avg - 1.0);
    *temperature = resistance_to_temperature(r);
    printk("ADC: %ld | RESISTANCE: %ld | TEMPERATURE: %d\r\n", adc_avg, (int32_t) r, *temperature);
    return err;
}

int ntc_init(void) {
    int err;

    adc_dev = device_get_binding(ADC_DEV);
    if (!adc_dev) {
        printk("Error getting " ADC_DEV " failed\n");
        return 1;
    }
    gpio_dev = device_get_binding(GPIO_DEV);
    if (!gpio_dev) {
        printk("Error getting " GPIO_DEV " failed\n");
        return 1;
    }
    static const struct adc_channel_cfg m_1st_channel_cfg = {
            .gain = ADC_GAIN,
            .reference = ADC_REFERENCE,
            .acquisition_time = ADC_ACQUISITION_TIME,
            .channel_id = ADC_1ST_CHANNEL_ID,
#if defined(CONFIG_ADC_CONFIGURABLE_INPUTS)
            .input_positive = ADC_1ST_CHANNEL_INPUT,
#endif
    };

    err = adc_channel_setup(adc_dev, &m_1st_channel_cfg);
    if (err) {
        printk("Error in adc setup: %d\n", err);
        return err;
    }

    err = gpio_pin_configure(gpio_dev, 28, GPIO_OUTPUT);
    if (err) {
        printk("Error in gpio setup: %d\n", err);
        return err;
    }
    gpio_pin_set(gpio_dev, 28, 1);
    return err;
}
